"""
collect cmake -LH output

for inclusion in docs
"""

import sys
from pathlib import Path
from subprocess import PIPE, run
from tempfile import TemporaryDirectory

here = Path(__file__).parent.absolute()
repo = here.parent
home = str(Path.home())


def summarize_cmake_output(text: str) -> str:
    """Summarize cmake -LH output

    Formats help strings nicer, excludes common
    """
    text = text.replace(sys.prefix, "$PREFIX")
    text = text.replace(home, "~")
    chunks = text.split("\n\n")
    new_chunks = []
    for chunk in chunks:
        if not chunk:
            continue
        lines = chunk.splitlines()
        doc_lines, assignment = lines[:-1], lines[-1]
        if assignment.startswith(("CMAKE_", "FETCHCONTENT_")):
            continue
        doc_lines = [
            "# " + doc_line.lstrip("/ ")
            for doc_line in doc_lines
            if not doc_line.startswith("--")
        ]
        new_chunks.append("\n".join(doc_lines + [assignment]))
    return "\n\n".join(new_chunks)


def summarize_cmake(path: Path) -> str:
    """Collect summarized cmake -LH output from a repo"""
    path = Path(path).absolute()
    with TemporaryDirectory() as td:
        p = run(
            ["cmake", "-LH", str(path)],
            text=True,
            stderr=sys.stderr,
            stdout=PIPE,
            check=False,
            cwd=td,
        )
    return summarize_cmake_output(p.stdout)


def main():
    if len(sys.argv) < 2:
        paths = [repo]
    else:
        paths = sys.argv[1:]
    for path in paths:
        print(path)
        print(summarize_cmake(path))
        print("\n\n")


if __name__ == "__main__":
    main()



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import time

import zmq
from zmq import devices
from zmq_test_utils import PYPY, BaseZMQTestCase, GreenTest, SkipTest, have_gevent

if PYPY:
    # cleanup of shared Context doesn't work on PyPy
    devices.Device.context_factory = zmq.Context


class TestDevice(BaseZMQTestCase):
    def test_device_types(self):
        for devtype in (zmq.STREAMER, zmq.FORWARDER, zmq.QUEUE):
            dev = devices.Device(devtype, zmq.PAIR, zmq.PAIR)
            assert dev.device_type == devtype
            del dev

    def test_device_attributes(self):
        dev = devices.Device(zmq.QUEUE, zmq.SUB, zmq.PUB)
        assert dev.in_type == zmq.SUB
        assert dev.out_type == zmq.PUB
        assert dev.device_type == zmq.QUEUE
        assert dev.daemon is True
        del dev

    def test_single_socket_forwarder_connect(self):
        if zmq.zmq_version() in ('4.1.1', '4.0.6'):
            raise SkipTest(f"libzmq-{zmq.zmq_version()} broke single-socket devices")
        dev = devices.ThreadDevice(zmq.QUEUE, zmq.REP, -1)
        req = self.context.socket(zmq.REQ)
        port = req.bind_to_random_port('tcp://127.0.0.1')
        dev.connect_in('tcp://127.0.0.1:%i' % port)
        dev.start()
        time.sleep(0.25)
        msg = b'hello'
        req.send(msg)
        assert msg == self.recv(req)
        del dev
        req.close()
        dev = devices.ThreadDevice(zmq.QUEUE, zmq.REP, -1)
        req = self.context.socket(zmq.REQ)
        port = req.bind_to_random_port('tcp://127.0.0.1')
        dev.connect_out('tcp://127.0.0.1:%i' % port)
        dev.start()
        time.sleep(0.25)
        msg = b'hello again'
        req.send(msg)
        assert msg == self.recv(req)
        del dev
        req.close()

    def test_single_socket_forwarder_bind(self):
        if zmq.zmq_version() in ('4.1.1', '4.0.6'):
            raise SkipTest(f"libzmq-{zmq.zmq_version()} broke single-socket devices")
        dev = devices.ThreadDevice(zmq.QUEUE, zmq.REP, -1)
        port = dev.bind_in_to_random_port('tcp://127.0.0.1')
        req = self.context.socket(zmq.REQ)
        req.connect('tcp://127.0.0.1:%i' % port)
        dev.start()
        time.sleep(0.25)
        msg = b'hello'
        req.send(msg)
        assert msg == self.recv(req)
        del dev
        req.close()
        dev = devices.ThreadDevice(zmq.QUEUE, zmq.REP, -1)
        port = dev.bind_in_to_random_port('tcp://127.0.0.1')
        req = self.context.socket(zmq.REQ)
        req.connect('tcp://127.0.0.1:%i' % port)
        dev.start()
        time.sleep(0.25)
        msg = b'hello again'
        req.send(msg)
        assert msg == self.recv(req)
        del dev
        req.close()

    def test_device_bind_to_random_with_args(self):
        dev = devices.ThreadDevice(zmq.PULL, zmq.PUSH, -1)
        iface = 'tcp://127.0.0.1'
        ports = []
        min, max = 5000, 5050
        ports.extend(
            [
                dev.bind_in_to_random_port(iface, min_port=min, max_port=max),
                dev.bind_out_to_random_port(iface, min_port=min, max_port=max),
            ]
        )
        for port in ports:
            if port < min or port > max:
                self.fail('Unexpected port number: %i' % port)

    def test_device_bind_to_random_binderror(self):
        dev = devices.ThreadDevice(zmq.PULL, zmq.PUSH, -1)
        iface = 'tcp://127.0.0.1'
        try:
            for i in range(11):
                dev.bind_in_to_random_port(iface, min_port=10000, max_port=10010)
        except zmq.ZMQBindError as e:
            return
        else:
            self.fail('Should have failed')

    def test_proxy(self):
        if zmq.zmq_version_info() < (3, 2):
            raise SkipTest("Proxies only in libzmq >= 3")
        dev = devices.ThreadProxy(zmq.PULL, zmq.PUSH, zmq.PUSH)
        iface = 'tcp://127.0.0.1'
        port = dev.bind_in_to_random_port(iface)
        port2 = dev.bind_out_to_random_port(iface)
        port3 = dev.bind_mon_to_random_port(iface)
        dev.start()
        time.sleep(0.25)
        msg = b'hello'
        push = self.context.socket(zmq.PUSH)
        push.connect("%s:%i" % (iface, port))
        pull = self.context.socket(zmq.PULL)
        pull.connect("%s:%i" % (iface, port2))
        mon = self.context.socket(zmq.PULL)
        mon.connect("%s:%i" % (iface, port3))
        push.send(msg)
        self.sockets.extend([push, pull, mon])
        assert msg == self.recv(pull)
        assert msg == self.recv(mon)

    def test_proxy_bind_to_random_with_args(self):
        if zmq.zmq_version_info() < (3, 2):
            raise SkipTest("Proxies only in libzmq >= 3")
        dev = devices.ThreadProxy(zmq.PULL, zmq.PUSH, zmq.PUSH)
        iface = 'tcp://127.0.0.1'
        ports = []
        min, max = 5000, 5050
        ports.extend(
            [
                dev.bind_in_to_random_port(iface, min_port=min, max_port=max),
                dev.bind_out_to_random_port(iface, min_port=min, max_port=max),
                dev.bind_mon_to_random_port(iface, min_port=min, max_port=max),
            ]
        )
        for port in ports:
            if port < min or port > max:
                self.fail('Unexpected port number: %i' % port)


if have_gevent:
    import gevent

    import zmq.green

    class TestDeviceGreen(GreenTest, BaseZMQTestCase):
        def test_green_device(self):
            rep = self.context.socket(zmq.REP)
            req = self.context.socket(zmq.REQ)
            self.sockets.extend([req, rep])
            port = rep.bind_to_random_port('tcp://127.0.0.1')
            g = gevent.spawn(zmq.green.device, zmq.QUEUE, rep, rep)
            req.connect('tcp://127.0.0.1:%i' % port)
            req.send(b'hi')
            timeout = gevent.Timeout(3)
            timeout.start()
            receiver = gevent.spawn(req.recv)
            assert receiver.get(2) == b'hi'
            timeout.cancel()
            g.kill(block=True)



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

from threading import Thread

import zmq
from zmq import Again, ContextTerminated, ZMQError, strerror
from zmq_test_utils import BaseZMQTestCase


class TestZMQError(BaseZMQTestCase):
    def test_strerror(self):
        """test that strerror gets the right type."""
        for i in range(10):
            e = strerror(i)
            assert isinstance(e, str)

    def test_zmqerror(self):
        for errno in range(10):
            e = ZMQError(errno)
            assert e.errno == errno
            assert str(e) == strerror(errno)

    def test_again(self):
        s = self.context.socket(zmq.REP)
        self.assertRaises(Again, s.recv, zmq.NOBLOCK)
        self.assertRaisesErrno(zmq.EAGAIN, s.recv, zmq.NOBLOCK)
        s.close()

    def atest_ctxterm(self):
        s = self.context.socket(zmq.REP)
        t = Thread(target=self.context.term)
        t.start()
        self.assertRaises(ContextTerminated, s.recv, zmq.NOBLOCK)
        self.assertRaisesErrno(zmq.TERM, s.recv, zmq.NOBLOCK)
        s.close()
        t.join()



"""
Locate the latest MSVC redist dir

and add it to $GITHUB_PATH so delvewheel can find the DLLs

finds 'C:/Program Files/Microsoft Visual Studio/2022/Enterprise/VC/Redist/MSVC/14.38.33135/arm64/Microsoft.VC143.CRT'
as of writing (2024-02-27)
"""

import os
import sys
from pathlib import Path


def log(msg):
    """Log a message to stderr"""
    print(msg, file=sys.stderr)


vs_version = "2022"
arch = os.environ.get("CIBW_ARCHS", "arm64")
vc_redist_path = (
    Path("C:/Program Files/Microsoft Visual Studio")
    / vs_version
    / "Enterprise/VC/Redist/MSVC"
)

log("Found VC redist versions:")
for v in vc_redist_path.glob("*"):
    log(v)


def _sort_key(dll_path):
    # redist paths look like
    # C:/.../MSVC/14.38.33135/
    # sort by the version number in the directory below MSVC
    version_dir = dll_path.relative_to(vc_redist_path).parents[-2]
    version_str = version_dir.name
    try:
        return tuple(int(part) for part in version_str.split("."))
    except ValueError:
        log(f"Not an apparent version: {version_str}")
        return (0, 0, 0, version_str)


log(f"Found msvcp for {arch}:")
# looking for .../MSVC/x.y.z/arm64/Microsoft.VC143.CRT/msvcp140.dll
# specifically *, not ** because we don't want onecore/arm64/...
found_arm_msvcp = sorted(
    vc_redist_path.glob(f"*/{arch}/**/msvcp140.dll"), key=_sort_key
)

for dll in found_arm_msvcp:
    log(dll)

selected_path = found_arm_msvcp[-1].parent
log(f"Selecting {selected_path}")

if os.environ.get("GITHUB_PATH"):
    log(f"Adding {selected_path} to $GITHUB_PATH")
    with open(os.environ["GITHUB_PATH"], "a") as f:
        f.write(str(selected_path) + "\n")



cdef extern from "zmq.h" nogil:
    enum: PYZMQ_DRAFT_API
    enum: ZMQ_VERSION
    enum: ZMQ_VERSION_MAJOR
    enum: ZMQ_VERSION_MINOR
    enum: ZMQ_VERSION_PATCH
    enum: ZMQ_IO_THREADS
    enum: ZMQ_MAX_SOCKETS
    enum: ZMQ_SOCKET_LIMIT
    enum: ZMQ_THREAD_PRIORITY
    enum: ZMQ_THREAD_SCHED_POLICY
    enum: ZMQ_MAX_MSGSZ
    enum: ZMQ_MSG_T_SIZE
    enum: ZMQ_THREAD_AFFINITY_CPU_ADD
    enum: ZMQ_THREAD_AFFINITY_CPU_REMOVE
    enum: ZMQ_THREAD_NAME_PREFIX
    enum: ZMQ_STREAMER
    enum: ZMQ_FORWARDER
    enum: ZMQ_QUEUE
    enum: ZMQ_EAGAIN "EAGAIN"
    enum: ZMQ_EFAULT "EFAULT"
    enum: ZMQ_EINVAL "EINVAL"
    enum: ZMQ_ENOTSUP "ENOTSUP"
    enum: ZMQ_EPROTONOSUPPORT "EPROTONOSUPPORT"
    enum: ZMQ_ENOBUFS "ENOBUFS"
    enum: ZMQ_ENETDOWN "ENETDOWN"
    enum: ZMQ_EADDRINUSE "EADDRINUSE"
    enum: ZMQ_EADDRNOTAVAIL "EADDRNOTAVAIL"
    enum: ZMQ_ECONNREFUSED "ECONNREFUSED"
    enum: ZMQ_EINPROGRESS "EINPROGRESS"
    enum: ZMQ_ENOTSOCK "ENOTSOCK"
    enum: ZMQ_EMSGSIZE "EMSGSIZE"
    enum: ZMQ_EAFNOSUPPORT "EAFNOSUPPORT"
    enum: ZMQ_ENETUNREACH "ENETUNREACH"
    enum: ZMQ_ECONNABORTED "ECONNABORTED"
    enum: ZMQ_ECONNRESET "ECONNRESET"
    enum: ZMQ_ENOTCONN "ENOTCONN"
    enum: ZMQ_ETIMEDOUT "ETIMEDOUT"
    enum: ZMQ_EHOSTUNREACH "EHOSTUNREACH"
    enum: ZMQ_ENETRESET "ENETRESET"
    enum: ZMQ_EFSM "EFSM"
    enum: ZMQ_ENOCOMPATPROTO "ENOCOMPATPROTO"
    enum: ZMQ_ETERM "ETERM"
    enum: ZMQ_EMTHREAD "EMTHREAD"
    enum: ZMQ_PROTOCOL_ERROR_WS_UNSPECIFIED
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_UNSPECIFIED
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_UNEXPECTED_COMMAND
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_INVALID_SEQUENCE
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_KEY_EXCHANGE
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_UNSPECIFIED
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_MESSAGE
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_HELLO
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_INITIATE
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_ERROR
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_READY
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_WELCOME
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_INVALID_METADATA
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC
    enum: ZMQ_PROTOCOL_ERROR_ZMTP_MECHANISM_MISMATCH
    enum: ZMQ_PROTOCOL_ERROR_ZAP_UNSPECIFIED
    enum: ZMQ_PROTOCOL_ERROR_ZAP_MALFORMED_REPLY
    enum: ZMQ_PROTOCOL_ERROR_ZAP_BAD_REQUEST_ID
    enum: ZMQ_PROTOCOL_ERROR_ZAP_BAD_VERSION
    enum: ZMQ_PROTOCOL_ERROR_ZAP_INVALID_STATUS_CODE
    enum: ZMQ_PROTOCOL_ERROR_ZAP_INVALID_METADATA
    enum: ZMQ_EVENT_CONNECTED
    enum: ZMQ_EVENT_CONNECT_DELAYED
    enum: ZMQ_EVENT_CONNECT_RETRIED
    enum: ZMQ_EVENT_LISTENING
    enum: ZMQ_EVENT_BIND_FAILED
    enum: ZMQ_EVENT_ACCEPTED
    enum: ZMQ_EVENT_ACCEPT_FAILED
    enum: ZMQ_EVENT_CLOSED
    enum: ZMQ_EVENT_CLOSE_FAILED
    enum: ZMQ_EVENT_DISCONNECTED
    enum: ZMQ_EVENT_MONITOR_STOPPED
    enum: ZMQ_EVENT_HANDSHAKE_FAILED_NO_DETAIL
    enum: ZMQ_EVENT_HANDSHAKE_SUCCEEDED
    enum: ZMQ_EVENT_HANDSHAKE_FAILED_PROTOCOL
    enum: ZMQ_EVENT_HANDSHAKE_FAILED_AUTH
    enum: ZMQ_EVENT_ALL_V1
    enum: ZMQ_EVENT_ALL
    enum: ZMQ_EVENT_PIPES_STATS
    enum: ZMQ_EVENT_ALL_V2
    enum: ZMQ_DONTWAIT
    enum: ZMQ_SNDMORE
    enum: ZMQ_NOBLOCK
    enum: ZMQ_MORE
    enum: ZMQ_SHARED
    enum: ZMQ_SRCFD
    enum: ZMQ_NORM_FIXED
    enum: ZMQ_NORM_CC
    enum: ZMQ_NORM_CCL
    enum: ZMQ_NORM_CCE
    enum: ZMQ_NORM_CCE_ECNONLY
    enum: ZMQ_POLLIN
    enum: ZMQ_POLLOUT
    enum: ZMQ_POLLERR
    enum: ZMQ_POLLPRI
    enum: ZMQ_RECONNECT_STOP_CONN_REFUSED
    enum: ZMQ_RECONNECT_STOP_HANDSHAKE_FAILED
    enum: ZMQ_RECONNECT_STOP_AFTER_DISCONNECT
    enum: ZMQ_NOTIFY_CONNECT
    enum: ZMQ_NOTIFY_DISCONNECT
    enum: ZMQ_NULL
    enum: ZMQ_PLAIN
    enum: ZMQ_CURVE
    enum: ZMQ_GSSAPI
    enum: ZMQ_HWM
    enum: ZMQ_AFFINITY
    enum: ZMQ_ROUTING_ID
    enum: ZMQ_SUBSCRIBE
    enum: ZMQ_UNSUBSCRIBE
    enum: ZMQ_RATE
    enum: ZMQ_RECOVERY_IVL
    enum: ZMQ_SNDBUF
    enum: ZMQ_RCVBUF
    enum: ZMQ_RCVMORE
    enum: ZMQ_FD
    enum: ZMQ_EVENTS
    enum: ZMQ_TYPE
    enum: ZMQ_LINGER
    enum: ZMQ_RECONNECT_IVL
    enum: ZMQ_BACKLOG
    enum: ZMQ_RECONNECT_IVL_MAX
    enum: ZMQ_MAXMSGSIZE
    enum: ZMQ_SNDHWM
    enum: ZMQ_RCVHWM
    enum: ZMQ_MULTICAST_HOPS
    enum: ZMQ_RCVTIMEO
    enum: ZMQ_SNDTIMEO
    enum: ZMQ_LAST_ENDPOINT
    enum: ZMQ_ROUTER_MANDATORY
    enum: ZMQ_TCP_KEEPALIVE
    enum: ZMQ_TCP_KEEPALIVE_CNT
    enum: ZMQ_TCP_KEEPALIVE_IDLE
    enum: ZMQ_TCP_KEEPALIVE_INTVL
    enum: ZMQ_IMMEDIATE
    enum: ZMQ_XPUB_VERBOSE
    enum: ZMQ_ROUTER_RAW
    enum: ZMQ_IPV6
    enum: ZMQ_MECHANISM
    enum: ZMQ_PLAIN_SERVER
    enum: ZMQ_PLAIN_USERNAME
    enum: ZMQ_PLAIN_PASSWORD
    enum: ZMQ_CURVE_SERVER
    enum: ZMQ_CURVE_PUBLICKEY
    enum: ZMQ_CURVE_SECRETKEY
    enum: ZMQ_CURVE_SERVERKEY
    enum: ZMQ_PROBE_ROUTER
    enum: ZMQ_REQ_CORRELATE
    enum: ZMQ_REQ_RELAXED
    enum: ZMQ_CONFLATE
    enum: ZMQ_ZAP_DOMAIN
    enum: ZMQ_ROUTER_HANDOVER
    enum: ZMQ_TOS
    enum: ZMQ_CONNECT_ROUTING_ID
    enum: ZMQ_GSSAPI_SERVER
    enum: ZMQ_GSSAPI_PRINCIPAL
    enum: ZMQ_GSSAPI_SERVICE_PRINCIPAL
    enum: ZMQ_GSSAPI_PLAINTEXT
    enum: ZMQ_HANDSHAKE_IVL
    enum: ZMQ_SOCKS_PROXY
    enum: ZMQ_XPUB_NODROP
    enum: ZMQ_BLOCKY
    enum: ZMQ_XPUB_MANUAL
    enum: ZMQ_XPUB_WELCOME_MSG
    enum: ZMQ_STREAM_NOTIFY
    enum: ZMQ_INVERT_MATCHING
    enum: ZMQ_HEARTBEAT_IVL
    enum: ZMQ_HEARTBEAT_TTL
    enum: ZMQ_HEARTBEAT_TIMEOUT
    enum: ZMQ_XPUB_VERBOSER
    enum: ZMQ_CONNECT_TIMEOUT
    enum: ZMQ_TCP_MAXRT
    enum: ZMQ_THREAD_SAFE
    enum: ZMQ_MULTICAST_MAXTPDU
    enum: ZMQ_VMCI_BUFFER_SIZE
    enum: ZMQ_VMCI_BUFFER_MIN_SIZE
    enum: ZMQ_VMCI_BUFFER_MAX_SIZE
    enum: ZMQ_VMCI_CONNECT_TIMEOUT
    enum: ZMQ_USE_FD
    enum: ZMQ_GSSAPI_PRINCIPAL_NAMETYPE
    enum: ZMQ_GSSAPI_SERVICE_PRINCIPAL_NAMETYPE
    enum: ZMQ_BINDTODEVICE
    enum: ZMQ_IDENTITY
    enum: ZMQ_CONNECT_RID
    enum: ZMQ_TCP_ACCEPT_FILTER
    enum: ZMQ_IPC_FILTER_PID
    enum: ZMQ_IPC_FILTER_UID
    enum: ZMQ_IPC_FILTER_GID
    enum: ZMQ_IPV4ONLY
    enum: ZMQ_DELAY_ATTACH_ON_CONNECT
    enum: ZMQ_FAIL_UNROUTABLE
    enum: ZMQ_ROUTER_BEHAVIOR
    enum: ZMQ_ZAP_ENFORCE_DOMAIN
    enum: ZMQ_LOOPBACK_FASTPATH
    enum: ZMQ_METADATA
    enum: ZMQ_MULTICAST_LOOP
    enum: ZMQ_ROUTER_NOTIFY
    enum: ZMQ_XPUB_MANUAL_LAST_VALUE
    enum: ZMQ_SOCKS_USERNAME
    enum: ZMQ_SOCKS_PASSWORD
    enum: ZMQ_IN_BATCH_SIZE
    enum: ZMQ_OUT_BATCH_SIZE
    enum: ZMQ_WSS_KEY_PEM
    enum: ZMQ_WSS_CERT_PEM
    enum: ZMQ_WSS_TRUST_PEM
    enum: ZMQ_WSS_HOSTNAME
    enum: ZMQ_WSS_TRUST_SYSTEM
    enum: ZMQ_ONLY_FIRST_SUBSCRIBE
    enum: ZMQ_RECONNECT_STOP
    enum: ZMQ_HELLO_MSG
    enum: ZMQ_DISCONNECT_MSG
    enum: ZMQ_PRIORITY
    enum: ZMQ_BUSY_POLL
    enum: ZMQ_HICCUP_MSG
    enum: ZMQ_XSUB_VERBOSE_UNSUBSCRIBE
    enum: ZMQ_TOPICS_COUNT
    enum: ZMQ_NORM_MODE
    enum: ZMQ_NORM_UNICAST_NACK
    enum: ZMQ_NORM_BUFFER_SIZE
    enum: ZMQ_NORM_SEGMENT_SIZE
    enum: ZMQ_NORM_BLOCK_SIZE
    enum: ZMQ_NORM_NUM_PARITY
    enum: ZMQ_NORM_NUM_AUTOPARITY
    enum: ZMQ_NORM_PUSH
    enum: ZMQ_PAIR
    enum: ZMQ_PUB
    enum: ZMQ_SUB
    enum: ZMQ_REQ
    enum: ZMQ_REP
    enum: ZMQ_DEALER
    enum: ZMQ_ROUTER
    enum: ZMQ_PULL
    enum: ZMQ_PUSH
    enum: ZMQ_XPUB
    enum: ZMQ_XSUB
    enum: ZMQ_STREAM
    enum: ZMQ_XREQ
    enum: ZMQ_XREP
    enum: ZMQ_SERVER
    enum: ZMQ_CLIENT
    enum: ZMQ_RADIO
    enum: ZMQ_DISH
    enum: ZMQ_GATHER
    enum: ZMQ_SCATTER
    enum: ZMQ_DGRAM
    enum: ZMQ_PEER
    enum: ZMQ_CHANNEL



"""tests for extending pyzmq"""

import zmq


class CustomSocket(zmq.Socket):
    custom_attr: int

    def __init__(self, context, socket_type, custom_attr: int = 0):
        super().__init__(context, socket_type)
        self.custom_attr = custom_attr


class CustomContext(zmq.Context):
    extra_arg: str
    _socket_class = CustomSocket

    def __init__(self, extra_arg: str = 'x'):
        super().__init__()
        self.extra_arg = extra_arg


def test_custom_context():
    ctx = CustomContext('s')
    assert isinstance(ctx, CustomContext)

    assert ctx.extra_arg == 's'
    s = ctx.socket(zmq.PUSH, custom_attr=10)
    assert isinstance(s, CustomSocket)
    assert s.custom_attr == 10
    assert s.context is ctx
    assert s.type == zmq.PUSH
    s.close()
    ctx.term()



# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import zmq
import zmq.asyncio
from zmq.utils.monitor import recv_monitor_message
from zmq_test_utils import require_zmq_4

pytestmark = require_zmq_4
import pytest


@pytest.fixture(params=["zmq", "asyncio"])
def Context(request, event_loop):
    if request.param == "asyncio":
        return zmq.asyncio.Context
    else:
        return zmq.Context


async def test_monitor(context, socket):
    """Test monitoring interface for sockets."""
    s_rep = socket(zmq.REP)
    s_req = socket(zmq.REQ)
    s_req.bind("tcp://127.0.0.1:6666")
    # try monitoring the REP socket
    s_rep.monitor(
        "inproc://monitor.rep",
        zmq.EVENT_CONNECT_DELAYED | zmq.EVENT_CONNECTED | zmq.EVENT_MONITOR_STOPPED,
    )
    # create listening socket for monitor
    s_event = socket(zmq.PAIR)
    s_event.connect("inproc://monitor.rep")
    s_event.linger = 0
    # test receive event for connect event
    s_rep.connect("tcp://127.0.0.1:6666")
    m = recv_monitor_message(s_event)
    if isinstance(context, zmq.asyncio.Context):
        m = await m
    if m['event'] == zmq.EVENT_CONNECT_DELAYED:
        assert m['endpoint'] == b"tcp://127.0.0.1:6666"
        # test receive event for connected event
        m = recv_monitor_message(s_event)
        if isinstance(context, zmq.asyncio.Context):
            m = await m
    assert m['event'] == zmq.EVENT_CONNECTED
    assert m['endpoint'] == b"tcp://127.0.0.1:6666"

    # test monitor can be disabled.
    s_rep.disable_monitor()
    m = recv_monitor_message(s_event)
    if isinstance(context, zmq.asyncio.Context):
        m = await m
    assert m['event'] == zmq.EVENT_MONITOR_STOPPED


async def test_monitor_repeat(context, socket, sockets):
    s = socket(zmq.PULL)
    m = s.get_monitor_socket()
    sockets.append(m)
    m2 = s.get_monitor_socket()
    assert m is m2
    s.disable_monitor()
    evt = recv_monitor_message(m)
    if isinstance(context, zmq.asyncio.Context):
        evt = await evt
    assert evt['event'] == zmq.EVENT_MONITOR_STOPPED
    m.close()
    s.close()


async def test_monitor_connected(context, socket, sockets):
    """Test connected monitoring socket."""
    s_rep = socket(zmq.REP)
    s_req = socket(zmq.REQ)
    s_req.bind("tcp://127.0.0.1:6667")
    # try monitoring the REP socket
    # create listening socket for monitor
    s_event = s_rep.get_monitor_socket()
    s_event.linger = 0
    sockets.append(s_event)
    # test receive event for connect event
    s_rep.connect("tcp://127.0.0.1:6667")
    m = recv_monitor_message(s_event)
    if isinstance(context, zmq.asyncio.Context):
        m = await m
    if m['event'] == zmq.EVENT_CONNECT_DELAYED:
        assert m['endpoint'] == b"tcp://127.0.0.1:6667"
        # test receive event for connected event
        m = recv_monitor_message(s_event)
        if isinstance(context, zmq.asyncio.Context):
            m = await m
    assert m['event'] == zmq.EVENT_CONNECTED
    assert m['endpoint'] == b"tcp://127.0.0.1:6667"



"""Future-returning APIs for coroutines."""

# Copyright (c) PyZMQ Developers.
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import warnings
from asyncio import Future
from collections import deque
from functools import partial
from itertools import chain
from typing import Any, Awaitable, Callable, NamedTuple, TypeVar, cast, overload

import zmq as _zmq
from zmq import EVENTS, POLLIN, POLLOUT
from zmq._typing import Literal


class _FutureEvent(NamedTuple):
    future: Future
    kind: str
    kwargs: dict
    msg: Any
    timer: Any


# These are incomplete classes and need a Mixin for compatibility with an eventloop
# defining the following attributes:
#
# _Future
# _READ
# _WRITE
# _default_loop()


class _Async:
    """Mixin for common async logic"""

    _current_loop: Any = None
    _Future: type[Future]

    def _get_loop(self) -> Any:
        """Get event loop

        Notice if event loop has changed,
        and register init_io_state on activation of a new event loop
        """
        if self._current_loop is None:
            self._current_loop = self._default_loop()
            self._init_io_state(self._current_loop)
            return self._current_loop
        current_loop = self._default_loop()
        if current_loop is not self._current_loop:
            # warn? This means a socket is being used in multiple loops!
            self._current_loop = current_loop
            self._init_io_state(current_loop)
        return current_loop

    def _default_loop(self) -> Any:
        raise NotImplementedError("Must be implemented in a subclass")

    def _init_io_state(self, loop=None) -> None:
        pass


class _AsyncPoller(_Async, _zmq.Poller):
    """Poller that returns a Future on poll, instead of blocking."""

    _socket_class: type[_AsyncSocket]
    _READ: int
    _WRITE: int
    raw_sockets: list[Any]

    def _watch_raw_socket(self, loop: Any, socket: Any, evt: int, f: Callable) -> None:
        """Schedule callback for a raw socket"""
        raise NotImplementedError()

    def _unwatch_raw_sockets(self, loop: Any, *sockets: Any) -> None:
        """Unschedule callback for a raw socket"""
        raise NotImplementedError()

    def poll(self, timeout=-1) -> Awaitable[list[tuple[Any, int]]]:  # type: ignore
        """Return a Future for a poll event"""
        future = self._Future()
        if timeout == 0:
            try:
                result = super().poll(0)
            except Exception as e:
                future.set_exception(e)
            else:
                future.set_result(result)
            return future

        loop = self._get_loop()

        # register Future to be called as soon as any event is available on any socket
        watcher = self._Future()

        # watch raw sockets:
        raw_sockets: list[Any] = []

        def wake_raw(*args):
            if not watcher.done():
                watcher.set_result(None)

        watcher.add_done_callback(
            lambda f: self._unwatch_raw_sockets(loop, *raw_sockets)
        )

        wrapped_sockets: list[_AsyncSocket] = []

        def _clear_wrapper_io(f):
            for s in wrapped_sockets:
                s._clear_io_state()

        for socket, mask in self.sockets:
            if isinstance(socket, _zmq.Socket):
                if not isinstance(socket, self._socket_class):
                    # it's a blocking zmq.Socket, wrap it in async
                    socket = self._socket_class.from_socket(socket)
                    wrapped_sockets.append(socket)
                if mask & _zmq.POLLIN:
                    socket._add_recv_event('poll', future=watcher)
                if mask & _zmq.POLLOUT:
                    socket._add_send_event('poll', future=watcher)
            else:
                raw_sockets.append(socket)
                evt = 0
                if mask & _zmq.POLLIN:
                    evt |= self._READ
                if mask & _zmq.POLLOUT:
                    evt |= self._WRITE
                self._watch_raw_socket(loop, socket, evt, wake_raw)

        def on_poll_ready(f):
            if future.done():
                return
            if watcher.cancelled():
                try:
                    future.cancel()
                except RuntimeError:
                    # RuntimeError may be called during teardown
                    pass
                return
            if watcher.exception():
                future.set_exception(watcher.exception())
            else:
                try:
                    result = super(_AsyncPoller, self).poll(0)
                except Exception as e:
                    future.set_exception(e)
                else:
                    future.set_result(result)

        watcher.add_done_callback(on_poll_ready)

        if wrapped_sockets:
            watcher.add_done_callback(_clear_wrapper_io)

        if timeout is not None and timeout > 0:
            # schedule cancel to fire on poll timeout, if any
            def trigger_timeout():
                if not watcher.done():
                    watcher.set_result(None)

            timeout_handle = loop.call_later(1e-3 * timeout, trigger_timeout)

            def cancel_timeout(f):
                if hasattr(timeout_handle, 'cancel'):
                    timeout_handle.cancel()
                else:
                    loop.remove_timeout(timeout_handle)

            future.add_done_callback(cancel_timeout)

        def cancel_watcher(f):
            if not watcher.done():
                watcher.cancel()

        future.add_done_callback(cancel_watcher)

        return future


class _NoTimer:
    @staticmethod
    def cancel():
        pass


T = TypeVar("T", bound="_AsyncSocket")


class _AsyncSocket(_Async, _zmq.Socket[Future]):
    # Warning : these class variables are only here to allow to call super().__setattr__.
    # They be overridden at instance initialization and not shared in the whole class
    _recv_futures = None
    _send_futures = None
    _state = 0
    _shadow_sock: _zmq.Socket
    _poller_class = _AsyncPoller
    _fd = None

    def __init__(
        self,
        context=None,
        socket_type=-1,
        io_loop=None,
        _from_socket: _zmq.Socket | None = None,
        **kwargs,
    ) -> None:
        if isinstance(context, _zmq.Socket):
            context, _from_socket = (None, context)
        if _from_socket is not None:
            super().__init__(shadow=_from_socket.underlying)  # type: ignore
            self._shadow_sock = _from_socket
        else:
            super().__init__(context, socket_type, **kwargs)  # type: ignore
            self._shadow_sock = _zmq.Socket.shadow(self.underlying)

        if io_loop is not None:
            warnings.warn(
                f"{self.__class__.__name__}(io_loop) argument is deprecated in pyzmq 22.2."
                " The currently active loop will always be used.",
                DeprecationWarning,
                stacklevel=3,
            )
        self._recv_futures = deque()
        self._send_futures = deque()
        self._state = 0
        self._fd = self._shadow_sock.FD

    @classmethod
    def from_socket(cls: type[T], socket: _zmq.Socket, io_loop: Any = None) -> T:
        """Create an async socket from an existing Socket"""
        return cls(_from_socket=socket, io_loop=io_loop)

    def close(self, linger: int | None = None) -> None:
        if not self.closed and self._fd is not None:
            event_list: list[_FutureEvent] = list(
                chain(self._recv_futures or [], self._send_futures or [])
            )
            for event in event_list:
                if not event.future.done():
                    try:
                        event.future.cancel()
                    except RuntimeError:
                        # RuntimeError may be called during teardown
                        pass
            self._clear_io_state()
        super().close(linger=linger)

    close.__doc__ = _zmq.Socket.close.__doc__

    def get(self, key):
        result = super().get(key)
        if key == EVENTS:
            self._schedule_remaining_events(result)
        return result

    get.__doc__ = _zmq.Socket.get.__doc__

    @overload  # type: ignore
    def recv_multipart(
        self, flags: int = 0, *, track: bool = False
    ) -> Awaitable[list[bytes]]: ...

    @overload
    def recv_multipart(
        self, flags: int = 0, *, copy: Literal[True], track: bool = False
    ) -> Awaitable[list[bytes]]: ...

    @overload
    def recv_multipart(
        self, flags: int = 0, *, copy: Literal[False], track: bool = False
    ) -> Awaitable[list[_zmq.Frame]]:  # type: ignore
        ...

    @overload
    def recv_multipart(
        self, flags: int = 0, copy: bool = True, track: bool = False
    ) -> Awaitable[list[bytes] | list[_zmq.Frame]]: ...

    def recv_multipart(
        self, flags: int = 0, copy: bool = True, track: bool = False
    ) -> Awaitable[list[bytes] | list[_zmq.Frame]]:
        """Receive a complete multipart zmq message.

        Returns a Future whose result will be a multipart message.
        """
        return self._add_recv_event(
            'recv_multipart', dict(flags=flags, copy=copy, track=track)
        )

    @overload  # type: ignore
    def recv(self, flags: int = 0, *, track: bool = False) -> Awaitable[bytes]: ...

    @overload
    def recv(
        self, flags: int = 0, *, copy: Literal[True], track: bool = False
    ) -> Awaitable[bytes]: ...

    @overload
    def recv(
        self, flags: int = 0, *, copy: Literal[False], track: bool = False
    ) -> Awaitable[_zmq.Frame]: ...

    def recv(  # type: ignore
        self, flags: int = 0, copy: bool = True, track: bool = False
    ) -> Awaitable[bytes | _zmq.Frame]:
        """Receive a single zmq frame.

        Returns a Future, whose result will be the received frame.

        Recommend using recv_multipart instead.
        """
        return self._add_recv_event('recv', dict(flags=flags, copy=copy, track=track))

    def send_multipart(  # type: ignore
        self, msg_parts: Any, flags: int = 0, copy: bool = True, track=False, **kwargs
    ) -> Awaitable[_zmq.MessageTracker | None]:
        """Send a complete multipart zmq message.

        Returns a Future that resolves when sending is complete.
        """
        kwargs['flags'] = flags
        kwargs['copy'] = copy
        kwargs['track'] = track
        return self._add_send_event('send_multipart', msg=msg_parts, kwargs=kwargs)

    def send(  # type: ignore
        self,
        data: Any,
        flags: int = 0,
        copy: bool = True,
        track: bool = False,
        **kwargs: Any,
    ) -> Awaitable[_zmq.MessageTracker | None]:
        """Send a single zmq frame.

        Returns a Future that resolves when sending is complete.

        Recommend using send_multipart instead.
        """
        kwargs['flags'] = flags
        kwargs['copy'] = copy
        kwargs['track'] = track
        kwargs.update(dict(flags=flags, copy=copy, track=track))
        return self._add_send_event('send', msg=data, kwargs=kwargs)

    def _deserialize(self, recvd, load):
        """Deserialize with Futures"""
        f = self._Future()

        def _chain(_):
            """Chain result through serialization to recvd"""
            if f.done():
                # chained future may be cancelled, which means nobody is going to get this result
                # if it's an error, that's no big deal (probably zmq.Again),
                # but if it's a successful recv, this is a dropped message!
                if not recvd.cancelled() and recvd.exception() is None:
                    warnings.warn(
                        # is there a useful stacklevel?
                        # ideally, it would point to where `f.cancel()` was called
                        f"Future {f} completed while awaiting {recvd}. A message has been dropped!",
                        RuntimeWarning,
                    )
                return
            if recvd.exception():
                f.set_exception(recvd.exception())
            else:
                buf = recvd.result()
                try:
                    loaded = load(buf)
                except Exception as e:
                    f.set_exception(e)
                else:
                    f.set_result(loaded)

        recvd.add_done_callback(_chain)

        def _chain_cancel(_):
            """Chain cancellation from f to recvd"""
            if recvd.done():
                return
            if f.cancelled():
                recvd.cancel()

        f.add_done_callback(_chain_cancel)

        return f

    def poll(self, timeout=None, flags=_zmq.POLLIN) -> Awaitable[int]:  # type: ignore
        """poll the socket for events

        returns a Future for the poll results.
        """

        if self.closed:
            raise _zmq.ZMQError(_zmq.ENOTSUP)

        p = self._poller_class()
        p.register(self, flags)
        poll_future = cast(Future, p.poll(timeout))

        future = self._Future()

        def unwrap_result(f):
            if future.done():
                return
            if poll_future.cancelled():
                try:
                    future.cancel()
                except RuntimeError:
                    # RuntimeError may be called during teardown
                    pass
                return
            if f.exception():
                future.set_exception(poll_future.exception())
            else:
                evts = dict(poll_future.result())
                future.set_result(evts.get(self, 0))

        if poll_future.done():
            # hook up result if already done
            unwrap_result(poll_future)
        else:
            poll_future.add_done_callback(unwrap_result)

        def cancel_poll(future):
            """Cancel underlying poll if request has been cancelled"""
            if not poll_future.done():
                try:
                    poll_future.cancel()
                except RuntimeError:
                    # RuntimeError may be called during teardown
                    pass

        future.add_done_callback(cancel_poll)

        return future

    # overrides only necessary for updated types
    def recv_string(self, *args, **kwargs) -> Awaitable[str]:  # type: ignore
        return super().recv_string(*args, **kwargs)  # type: ignore

    def send_string(  # type: ignore
        self, s: str, flags: int = 0, encoding: str = 'utf-8'
    ) -> Awaitable[None]:
        return super().send_string(s, flags=flags, encoding=encoding)  # type: ignore

    def _add_timeout(self, future, timeout):
        """Add a timeout for a send or recv Future"""

        def future_timeout():
            if future.done():
                # future already resolved, do nothing
                return

            # raise EAGAIN
            future.set_exception(_zmq.Again())

        return self._call_later(timeout, future_timeout)

    def _call_later(self, delay, callback):
        """Schedule a function to be called later

        Override for different IOLoop implementations

        Tornado and asyncio happen to both have ioloop.call_later
        with the same signature.
        """
        return self._get_loop().call_later(delay, callback)

    @staticmethod
    def _remove_finished_future(future, event_list, event=None):
        """Make sure that futures are removed from the event list when they resolve

        Avoids delaying cleanup until the next send/recv event,
        which may never come.
        """
        # "future" instance is shared between sockets, but each socket has its own event list.
        if not event_list:
            return
        # only unconsumed events (e.g. cancelled calls)
        # will be present when this happens
        try:
            event_list.remove(event)
        except ValueError:
            # usually this will have been removed by being consumed
            return

    def _add_recv_event(self, kind, kwargs=None, future=None):
        """Add a recv event, returning the corresponding Future"""
        f = future or self._Future()
        if kind.startswith('recv') and kwargs.get('flags', 0) & _zmq.DONTWAIT:
            # short-circuit non-blocking calls
            recv = getattr(self._shadow_sock, kind)
            try:
                r = recv(**kwargs)
            except Exception as e:
                f.set_exception(e)
            else:
                f.set_result(r)
            return f

        timer = _NoTimer
        if hasattr(_zmq, 'RCVTIMEO'):
            timeout_ms = self._shadow_sock.rcvtimeo
            if timeout_ms >= 0:
                timer = self._add_timeout(f, timeout_ms * 1e-3)

        # we add it to the list of futures before we add the timeout as the
        # timeout will remove the future from recv_futures to avoid leaks
        _future_event = _FutureEvent(f, kind, kwargs, msg=None, timer=timer)
        self._recv_futures.append(_future_event)

        if self._shadow_sock.get(EVENTS) & POLLIN:
            # recv immediately, if we can
            self._handle_recv()
        if self._recv_futures and _future_event in self._recv_futures:
            # Don't let the Future sit in _recv_events after it's done
            # no need to register this if we've already been handled
            # (i.e. immediately-resolved recv)
            f.add_done_callback(
                partial(
                    self._remove_finished_future,
                    event_list=self._recv_futures,
                    event=_future_event,
                )
            )
            self._add_io_state(POLLIN)
        return f

    def _add_send_event(self, kind, msg=None, kwargs=None, future=None):
        """Add a send event, returning the corresponding Future"""
        f = future or self._Future()
        # attempt send with DONTWAIT if no futures are waiting
        # short-circuit for sends that will resolve immediately
        # only call if no send Futures are waiting
        if kind in ('send', 'send_multipart') and not self._send_futures:
            flags = kwargs.get('flags', 0)
            nowait_kwargs = kwargs.copy()
            nowait_kwargs['flags'] = flags | _zmq.DONTWAIT

            # short-circuit non-blocking calls
            send = getattr(self._shadow_sock, kind)
            # track if the send resolved or not
            # (EAGAIN if DONTWAIT is not set should proceed with)
            finish_early = True
            try:
                r = send(msg, **nowait_kwargs)
            except _zmq.Again as e:
                if flags & _zmq.DONTWAIT:
                    f.set_exception(e)
                else:
                    # EAGAIN raised and DONTWAIT not requested,
                    # proceed with async send
                    finish_early = False
            except Exception as e:
                f.set_exception(e)
            else:
                f.set_result(r)

            if finish_early:
                # short-circuit resolved, return finished Future
                # schedule wake for recv if there are any receivers waiting
                if self._recv_futures:
                    self._schedule_remaining_events()
                return f

        timer = _NoTimer
        if hasattr(_zmq, 'SNDTIMEO'):
            timeout_ms = self._shadow_sock.get(_zmq.SNDTIMEO)
            if timeout_ms >= 0:
                timer = self._add_timeout(f, timeout_ms * 1e-3)

        # we add it to the list of futures before we add the timeout as the
        # timeout will remove the future from recv_futures to avoid leaks
        _future_event = _FutureEvent(f, kind, kwargs=kwargs, msg=msg, timer=timer)
        self._send_futures.append(_future_event)
        # Don't let the Future sit in _send_futures after it's done
        f.add_done_callback(
            partial(
                self._remove_finished_future,
                event_list=self._send_futures,
                event=_future_event,
            )
        )

        self._add_io_state(POLLOUT)
        return f

    def _handle_recv(self):
        """Handle recv events"""
        if not self._shadow_sock.get(EVENTS) & POLLIN:
            # event triggered, but state may have been changed between trigger and callback
            return
        f = None
        while self._recv_futures:
            f, kind, kwargs, _, timer = self._recv_futures.popleft()
            # skip any cancelled futures
            if f.done():
                f = None
            else:
                break

        if not self._recv_futures:
            self._drop_io_state(POLLIN)

        if f is None:
            return

        timer.cancel()

        if kind == 'poll':
            # on poll event, just signal ready, nothing else.
            f.set_result(None)
            return
        elif kind == 'recv_multipart':
            recv = self._shadow_sock.recv_multipart
        elif kind == 'recv':
            recv = self._shadow_sock.recv
        else:
            raise ValueError(f"Unhandled recv event type: {kind!r}")

        kwargs['flags'] |= _zmq.DONTWAIT
        try:
            result = recv(**kwargs)
        except Exception as e:
            f.set_exception(e)
        else:
            f.set_result(result)

    def _handle_send(self):
        if not self._shadow_sock.get(EVENTS) & POLLOUT:
            # event triggered, but state may have been changed between trigger and callback
            return
        f = None
        while self._send_futures:
            f, kind, kwargs, msg, timer = self._send_futures.popleft()
            # skip any cancelled futures
            if f.done():
                f = None
            else:
                break

        if not self._send_futures:
            self._drop_io_state(POLLOUT)

        if f is None:
            return

        timer.cancel()

        if kind == 'poll':
            # on poll event, just signal ready, nothing else.
            f.set_result(None)
            return
        elif kind == 'send_multipart':
            send = self._shadow_sock.send_multipart
        elif kind == 'send':
            send = self._shadow_sock.send
        else:
            raise ValueError(f"Unhandled send event type: {kind!r}")

        kwargs['flags'] |= _zmq.DONTWAIT
        try:
            result = send(msg, **kwargs)
        except Exception as e:
            f.set_exception(e)
        else:
            f.set_result(result)

    # event masking from ZMQStream
    def _handle_events(self, fd=0, events=0):
        """Dispatch IO events to _handle_recv, etc."""
        if self._shadow_sock.closed:
            return

        zmq_events = self._shadow_sock.get(EVENTS)
        if zmq_events & _zmq.POLLIN:
            self._handle_recv()
        if zmq_events & _zmq.POLLOUT:
            self._handle_send()
        self._schedule_remaining_events()

    def _schedule_remaining_events(self, events=None):
        """Schedule a call to handle_events next loop iteration

        If there are still events to handle.
        """
        # edge-triggered handling
        # allow passing events in, in case this is triggered by retrieving events,
        # so we don't have to retrieve it twice.
        if self._state == 0:
            # not watching for anything, nothing to schedule
            return
        if events is None:
            events = self._shadow_sock.get(EVENTS)
        if events & self._state:
            self._call_later(0, self._handle_events)

    def _add_io_state(self, state):
        """Add io_state to poller."""
        if self._state != state:
            state = self._state = self._state | state
        self._update_handler(self._state)

    def _drop_io_state(self, state):
        """Stop poller from watching an io_state."""
        if self._state & state:
            self._state = self._state & (~state)
        self._update_handler(self._state)

    def _update_handler(self, state):
        """Update IOLoop handler with state.

        zmq FD is always read-only.
        """
        # ensure loop is registered and init_io has been called
        # if there are any events to watch for
        if state:
            self._get_loop()
        self._schedule_remaining_events()

    def _init_io_state(self, loop=None):
        """initialize the ioloop event handler"""
        if loop is None:
            loop = self._get_loop()
        loop.add_handler(self._shadow_sock, self._handle_events, self._READ)
        self._call_later(0, self._handle_events)

    def _clear_io_state(self):
        """unregister the ioloop event handler

        called once during close
        """
        fd = self._shadow_sock
        if self._shadow_sock.closed:
            fd = self._fd
        if self._current_loop is not None:
            self._current_loop.remove_handler(fd)



"""
ROUTER-ROUTER communication example

ideal for P2P applications that need a socket that is able to connect
to several other peers simultaneously, while also being able
to receive messages

In this example, aiowire.EventLoop is used to await all tasks launched
within it, while timing out after (here) 1 second.
This guarantees completion of tasks - even if they are infinite loops
- as long as they regularly call await.

Contributed by github:jcpinto54 and github:frobnitzem
"""

# Copyright (c) 2022 PyZMQ Developers.
# This example is in the public domain (CC-0)

import asyncio

from aiowire import EventLoop

import zmq
from zmq.asyncio import Context


class Server:
    def __init__(self, url: str):
        context = Context.instance()
        socket = context.socket(zmq.ROUTER)
        socket.setsockopt_string(zmq.IDENTITY, 'server')
        socket.bind(url)
        self.socket = socket

    async def run(self, ev: EventLoop):
        # `Wire` type receives the event loop as its single argument

        req = await self.socket.recv_multipart()
        print(f"Server received {req}")
        # identity of receptionist, empty frame, message content
        await self.socket.send_multipart([req[0], b'', b'whatup'])

        # continue to run by returning this `Wire`
        return self.run


class Client:
    def __init__(self, url: str, name: str):
        self.name = name

        context = Context.instance()
        socket = context.socket(zmq.ROUTER)
        socket.setsockopt_string(zmq.IDENTITY, name)
        socket.connect(url)
        self.socket = socket

    async def run(self, ev: EventLoop):
        await asyncio.sleep(0.1)
        # identity of receptionist, empty frame, message content
        await self.socket.send_multipart([b'server', b'', b'cheers'])
        rep = await self.socket.recv_multipart()
        print(f"{self.name} received {rep}")

        # continue to run by returning this `Wire`
        return self.run


async def main() -> None:
    url = 'inproc://test_zmq'
    srv = Server(url)
    romeo = Client(url, 'romeo')
    sierra = Client(url, 'sierra')

    async with EventLoop(1.0) as ev:
        ev.start(romeo.run)
        ev.start(sierra.run)
        ev.start(srv.run)


if __name__ == "__main__":
    asyncio.run(main())



"""Test libzmq security (libzmq >= 3.3.0)"""

# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.

import contextlib
import os
import time
from threading import Thread

import zmq
from zmq.utils import z85
from zmq_test_utils import PYPY, BaseZMQTestCase, SkipTest

USER = b"admin"
PASS = b"password"


class TestSecurity(BaseZMQTestCase):
    def setUp(self):
        if zmq.zmq_version_info() < (4, 0):
            raise SkipTest("security is new in libzmq 4.0")
        try:
            zmq.curve_keypair()
        except zmq.ZMQError:
            raise SkipTest("security requires libzmq to be built with CURVE support")
        super().setUp()

    def zap_handler(self):
        socket = self.context.socket(zmq.REP)
        socket.bind("inproc://zeromq.zap.01")
        try:
            msg = self.recv_multipart(socket)

            version, sequence, domain, address, identity, mechanism = msg[:6]
            if mechanism == b'PLAIN':
                username, password = msg[6:]
            elif mechanism == b'CURVE':
                msg[6]

            assert version == b"1.0"
            assert identity == b"IDENT"
            reply = [version, sequence]
            if (
                mechanism == b'CURVE'
                or (mechanism == b'PLAIN' and username == USER and password == PASS)
                or (mechanism == b'NULL')
            ):
                reply.extend(
                    [
                        b"200",
                        b"OK",
                        b"anonymous",
                        b"\5Hello\0\0\0\5World",
                    ]
                )
            else:
                reply.extend(
                    [
                        b"400",
                        b"Invalid username or password",
                        b"",
                        b"",
                    ]
                )
            socket.send_multipart(reply)
        finally:
            socket.close()

    @contextlib.contextmanager
    def zap(self):
        self.start_zap()
        time.sleep(0.5)  # allow time for the Thread to start
        try:
            yield
        finally:
            self.stop_zap()

    def start_zap(self):
        self.zap_thread = Thread(target=self.zap_handler)
        self.zap_thread.start()

    def stop_zap(self):
        self.zap_thread.join()

    def bounce(self, server, client, test_metadata=True):
        msg = [os.urandom(64), os.urandom(64)]
        client.send_multipart(msg)
        frames = self.recv_multipart(server, copy=False)
        recvd = list(map(lambda x: x.bytes, frames))

        try:
            if test_metadata and not PYPY:
                for frame in frames:
                    assert frame.get('User-Id') == 'anonymous'
                    assert frame.get('Hello') == 'World'
                    assert frame['Socket-Type'] == 'DEALER'
        except zmq.ZMQVersionError:
            pass

        assert recvd == msg
        server.send_multipart(recvd)
        msg2 = self.recv_multipart(client)
        assert msg2 == msg

    def test_null(self):
        """test NULL (default) security"""
        server = self.socket(zmq.DEALER)
        client = self.socket(zmq.DEALER)
        assert client.MECHANISM == zmq.NULL
        assert server.mechanism == zmq.NULL
        assert client.plain_server == 0
        assert server.plain_server == 0
        iface = 'tcp://127.0.0.1'
        port = server.bind_to_random_port(iface)
        client.connect("%s:%i" % (iface, port))
        self.bounce(server, client, False)

    def test_plain(self):
        """test PLAIN authentication"""
        server = self.socket(zmq.DEALER)
        server.identity = b'IDENT'
        client = self.socket(zmq.DEALER)
        assert client.plain_username == b''
        assert client.plain_password == b''
        client.plain_username = USER
        client.plain_password = PASS
        assert client.getsockopt(zmq.PLAIN_USERNAME) == USER
        assert client.getsockopt(zmq.PLAIN_PASSWORD) == PASS
        assert client.plain_server == 0
        assert server.plain_server == 0
        server.plain_server = True
        assert server.mechanism == zmq.PLAIN
        assert client.mechanism == zmq.PLAIN

        assert not client.plain_server
        assert server.plain_server

        with self.zap():
            iface = 'tcp://127.0.0.1'
            port = server.bind_to_random_port(iface)
            client.connect("%s:%i" % (iface, port))
            self.bounce(server, client)

    def skip_plain_inauth(self):
        """test PLAIN failed authentication"""
        server = self.socket(zmq.DEALER)
        server.identity = b'IDENT'
        client = self.socket(zmq.DEALER)
        self.sockets.extend([server, client])
        client.plain_username = USER
        client.plain_password = b'incorrect'
        server.plain_server = True
        assert server.mechanism == zmq.PLAIN
        assert client.mechanism == zmq.PLAIN

        with self.zap():
            iface = 'tcp://127.0.0.1'
            port = server.bind_to_random_port(iface)
            client.connect("%s:%i" % (iface, port))
            client.send(b'ping')
            server.rcvtimeo = 250
            self.assertRaisesErrno(zmq.EAGAIN, server.recv)

    def test_keypair(self):
        """test curve_keypair"""
        try:
            public, secret = zmq.curve_keypair()
        except zmq.ZMQError:
            raise SkipTest("CURVE unsupported")

        assert type(secret) == bytes
        assert type(public) == bytes
        assert len(secret) == 40
        assert len(public) == 40

        # verify that it is indeed Z85
        bsecret, bpublic = (z85.decode(key) for key in (public, secret))
        assert type(bsecret) == bytes
        assert type(bpublic) == bytes
        assert len(bsecret) == 32
        assert len(bpublic) == 32

    def test_curve_public(self):
        """test curve_public"""
        try:
            public, secret = zmq.curve_keypair()
        except zmq.ZMQError:
            raise SkipTest("CURVE unsupported")
        if zmq.zmq_version_info() < (4, 2):
            raise SkipTest("curve_public is new in libzmq 4.2")

        derived_public = zmq.curve_public(secret)

        assert type(derived_public) == bytes
        assert len(derived_public) == 40

        # verify that it is indeed Z85
        bpublic = z85.decode(derived_public)
        assert type(bpublic) == bytes
        assert len(bpublic) == 32

        # verify that it is equal to the known public key
        assert derived_public == public

    def test_curve(self):
        """test CURVE encryption"""
        server = self.socket(zmq.DEALER)
        server.identity = b'IDENT'
        client = self.socket(zmq.DEALER)
        self.sockets.extend([server, client])
        try:
            server.curve_server = True
        except zmq.ZMQError as e:
            # will raise EINVAL if no CURVE support
            if e.errno == zmq.EINVAL:
                raise SkipTest("CURVE unsupported")

        server_public, server_secret = zmq.curve_keypair()
        client_public, client_secret = zmq.curve_keypair()

        server.curve_secretkey = server_secret
        server.curve_publickey = server_public
        client.curve_serverkey = server_public
        client.curve_publickey = client_public
        client.curve_secretkey = client_secret

        assert server.mechanism == zmq.CURVE
        assert client.mechanism == zmq.CURVE

        assert server.get(zmq.CURVE_SERVER)
        assert not client.get(zmq.CURVE_SERVER)

        with self.zap():
            iface = 'tcp://127.0.0.1'
            port = server.bind_to_random_port(iface)
            client.connect("%s:%i" % (iface, port))
            self.bounce(server, client)
